local M = {}

local scenario = require("scenario/scenarios")
local helper = require("scenario/scenariohelper")

local playerName = "rla_player0"
local trafficName = "rla_traffic"
local barrierNames = {"rla_barrier0", "rla_barrier1"}
local failMessages = {busted = "Busted! Step out of the vehicle, now!", crashBusted = "Totaled... and busted!", crash = "Totaled! Ouch.", overTime = "You ran out of time! What a disappointment."}
local winMessages = {normal = "Yay, you made it! The managers saw your vehicle and will keep the doors open for you.", lawful = "Yay, you made it! Nicely done, you didn't upset any cops!", lawless = "Yay, you made it! You took care of those roadblocks like a rock star.", veryFast = "Yay, you made it! That was fast! I admire your driving skills.", ultraFast = "Yay, you made it! Wow, that was insanely fast! Some of the local racers were informed, and they'd like to recruit you. That's crazy.", almost = "Yay, you made it! You got there with only seconds to spare!", policeCar = "Yay, you made it! Great vehicle choice; that's one way to deal with traffic!", funnyCar = "Yay, you made it! Oh wow, with that vehicle? Looks like you really know how to get around! Bravo.", evilCar = "Yay, you made it! That was pretty evil, though. What kind of monster are you?"}
local outMessages = {roadClosed = "Hey, you can't go through the construction zone!", wrongWay = "Wrong way! Nice try, buddy.", hospital = "You went to the hospital? I hope you're alright..."}
local paintColors = {{{1, 1, 1, 1.4}}, {{0, 0, 0, 1.4}, {0.5, 0, 0, 1.4}, {0.33, 0.33, 0.33, 1.4}, {0.0, 0.1, 0.42, 1.4}, {0.3, 0.2, 0.12, 1.4}, {0.83, 0.28, 0.03, 1.4}}, {{0.0, 0.05, 0.25, 1.2}, {0.5, 0, 0, 1.2}, {0.18, 0.18, 0.18, 1.2}, {0.45, 0.28, 0.233, 1.3}, {1, 0, 0, 1.2}, {0.22, 0.37, 0.33, 1.3}}, {{0.9, 0.05, 0.04, 1.35}, {1, 0.3, 0, 1.3}, {0.95, 0.7, 0.1, 1.35}, {0.0, 0.2, 0.7, 1.3}, {0.15, 0.35, 0.6, 1.3}, {0.66, 0.655, 0.65, 1.3}}, {{0.65, 0.65, 0.65, 1.2}, {0.2, 0.3, 0.6, 1.2}, {0.68, 0.1, 0.11, 1.3}, {0.24, 0.48, 0.48, 1.2}, {0, 0.04, 0.13, 1.2}, {0.3, 0, 0, 1.2}}} -- traffic colors; follow the curly brackets
local zoneData = {{0, 3, 0}, {1, 3, 0}, {2, 3, 5}, {3, 3, 10}, {4, 3, 15}, {5, 3, 20}, {6, 3, 25}, {7, 3, 25}, {8, 3, 25}, {9, 3, 10}, {10, 4, 40}, {11, 4, 40}, {12, 4, 40}, {13, 5, 40}, {14, 4, 40}, {15, 7, 40}, {16, 7, 40}, {17, 7, 40}, {18, 3, 40}, {19, 3, 40}, {20, 3, 40}, {21, 3, 40}, {22, 3, 40}, {23, 3, 40}, {24, 3, 40}, {25, 3, 40}, {26, 3, 35}, {27, 5, 35}, {28, 3, 30}, {29, 3, 30}, {30, 3, 30}, {31, 3, 30}, {32, 3, 30}, {33, 3, 30}, {34, 3, 30}, {35, 3, 25}, {36, 3, 25}, {37, 3, 25}, {38, 3, 25}, {39, 4, 20}, {40, 5, 15}, {41, 6, 10}, {42, 7, 30}, {43, 7, 30}, {44, 7, 30}, {45, 7, 30}, {46, 7, 30}, {47, 7, 30}, {48, 7, 30}, {49, 7, 30}} -- zone data; ID, side offset, forwards probability
local quatAngle = {quatFromDir(vec3(0, 1, 0), vec3(0, 0, 1)), quatFromDir(vec3(1, 0, 0), vec3(0, 0, 1)), quatFromDir(vec3(0, -1, 0), vec3(0, 0, 1)), quatFromDir(vec3(-1, 0, 0), vec3(0, 0, 1))} -- angles; 0, 90, 180, 270

local zoneNames = {}
local player = {}
local traffic = {}
local barriers = {}

local queue = 0
local roadblockZone = 1
local speedLimit = 0
local mainTimer = 0
local hornHold = 0
local funMode = false
local roadblockMode = false

local timeLimit = 200 -- time limit, in seconds
local playerSmash = 50000 -- amount of damage on the player car that will fail the scenario
local policeSmash = 5000 -- amount of damage on a police car that will trigger roadblocks
local trafficSmash = 2000 -- amount of damage on a traffic car that will enable it to reinitialize when spawning
local minSpeed = 15 -- minimum speed limit for randomizer, in metres per second
local maxSpeed = 20 -- maximum speed limit for randomizer, in metres per second
local speedingSpeed = 28 -- excessive speed; can trigger police mode
local crawlSpeed = 5 -- sloth speed; can trigger police arrest
local zoneGap = 3 -- gap between current zone and traffic spawn zone; 3 is optimal
local trafficAggro = 0.8 -- aggression level of the traffic cars
local nearDist = 20 -- near distance, used for traffic passing and interactions
local farDist = 320 -- far distance, used for traffic despawning
local readyDist = 100 -- normal distance, used for a variety of traffic and police logic
local farMultiplier = 3 -- spawn logic multipler for distant traffic cars; used with readyDist
local policeTime = 40 -- police car will be ready after this many seconds
local randomColors = false -- if true, traffic car colors can get changed to anything!

local function setSpawn(t) -- sets the new spawn point of a traffic car, then places and rotates it
	local allow = true
	local randomWay = math.random(1, 100)
	local zoneNumber = player.zone + zoneGap
	local zoneIn = scenetree.findObject(zoneNames[zoneNumber])
	local zonePos = zoneIn:getPosition()
	local zoneRot = zoneIn:getRotation()
	local objAngle = 0
	local posAngle = 0
	local shiftUp = vec3(0, 0, 0)
	local shiftOut = 0
	local pos = nil
	local rot = nil
	local colorIn = nil
	local colorValues = nil
	if t.role == "policeNormal" or (t.role == "policeChase" and roadblockMode == true) then -- normal police always spawn going forwards
		randomWay = 0
	end
	for i, v in ipairs(traffic) do
		if t.obj.name == v.obj.name then
			colorIn = paintColors[i][math.random(1, #paintColors[i])]
			shiftUp = vec3(0, 0, 0.5)
		end
		if v.zone == zoneNumber then
			allow = false	
		end
	end
	shiftOut = zoneData[zoneNumber][2]
	if randomWay <= zoneData[zoneNumber][3] then
		objAngle = 3
		posAngle = 4
		if t.role == "policeChase" and roadblockMode == true then -- police roadblock position
			objAngle = math.random(1, 2) * 2
			shiftOut = 2.8
		end
	else
		objAngle = 1
		posAngle = 2
	end
	if allow == true then
		if randomWay <= zoneData[zoneNumber][3] then
			t.travel = "forward"
		else
			t.travel = "backward"
		end
		pos = (vec3(zonePos) - vec3(0, 1, 0):rotated(zoneRot):rotated(quatAngle[posAngle]) * shiftOut) + shiftUp -- new position derived from spawn zone position
		rot = quatFromDir(vec3(0, 1, 0):rotated(zoneRot):rotated(quatAngle[objAngle]), vec3(0, 0, 1)) -- new rotation derived from spawn zone rotation
		t.obj:setPositionRotation(pos.x, pos.y, pos.z, 0, 0, rot.z, rot.w + (randomWay / 1000)) -- respawn point; the random addition fixes a spawn rotation bug
		t.obj:queueLuaCommand("vehicleResetted(false)")
		t.obj:resetBrokenFlexMesh()
		if t.role == "policeChase" and roadblockMode == true then
			t.role = "roadblock"
			roadblockZone = zoneNumber
		end
		if t.role == "policeChase" or t.role == "roadblock" then
			t.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
			player.policeWatch = "chase"
		end
		if randomColors == true then
			colorValues = {math.random(), math.random(), math.random(), math.random() + 1}
		else
			colorValues = {colorIn[1], colorIn[2], colorIn[3], colorIn[4]}
		end
		t.obj.color = ColorF(colorValues[1], colorValues[2], colorValues[3], colorValues[4]):asLinear4F()
		t.obj.colorPalette0 = t.obj.color
		t.state = "set"
		t.timer = 0
	end
end

local function setPropSpawn(p, shiftOut) -- sets the new spawn point of a prop, then places and rotates it
	if not shiftOut then shiftOut = 0 end
	local zoneNumber = roadblockZone
	local zoneIn = scenetree.findObject(zoneNames[zoneNumber])
	local zonePos = zoneIn:getPosition()
	local zoneRot = zoneIn:getRotation()
	local objAngle = 0
	local posAngle = 0
	local shiftUp = vec3(0, 0, 0)
	local pos = nil
	local rot = nil
	if p.role == "barrier" then
		objAngle = 1
		posAngle = 2
		shiftUp = vec3(0, 0, 0.3)
	end
	pos = (vec3(zonePos) - vec3(0, 1, 0):rotated(zoneRot):rotated(quatAngle[posAngle]) * shiftOut) + shiftUp
	rot = quatFromDir(vec3(0, 1, 0):rotated(zoneRot):rotated(quatAngle[objAngle]), vec3(0, 0, 1))
	p.obj:setPositionRotation(pos.x, pos.y, pos.z, 0, 0, rot.z, rot.w)
	p.obj:queueLuaCommand("vehicleResetted(false)")
	p.state = "set"
end

local function rushPlayer() -- gives the player a countdown to the time limit
	if player.timer % 1 == 0 then
		local countdown = 10 - math.modf(player.timer)
		if countdown > 0 then
			helper.flashUiMessage((tostring(countdown).."..."), 0.5, true)
		end
	end
	player.timer = player.timer + 0.25
	if player.timer > 10 then
		player.success = false
		scenario.finish({failed = failMessages.overTime})
	end
end

local function arrestPlayer() -- arrests the player with a countdown
	if player.arrestTimer % 1 == 0 then
		local countdown = 5 - math.modf(player.arrestTimer)
		if countdown > 0 then
			helper.flashUiMessage((tostring(countdown).."..."), 0.5, true)
		end
	end
	player.arrestTimer = player.arrestTimer + 0.25
	if player.arrestTimer > 5 then
		player.success = false
		scenario.finish({failed = failMessages.busted})
	end
end

local function chasePlayer(c) -- various logic for while the police car is chasing the player
	local avoid = false
	local chaseDist = readyDist / 2
	local playerDist = (player.pos - c.pos):len()
	if roadblockMode == true then -- police will spawn less or more often depending on these conditions
		c.far = farMultiplier / 1.5
	else
		if player.damage < trafficSmash then -- clean racers get less cops (better for time attacks)
			c.far = farMultiplier * 1.5
		else
			c.far = farMultiplier
		end
	end
	for i, v in ipairs(traffic) do
		if v.role == "normal" then
			v.pos = v.obj:getPosition()
			local policeDist = (c.pos - v.pos):len()
			if playerDist <= readyDist and policeDist <= chaseDist then -- traffic reacts to active AI police
				v.obj:queueLuaCommand("ai.setSpeed(5)")
				if c.travel == "forward" and v.travel == "forward" then -- AI police will try to dodge traffic
					avoid = true
				end
			end
		end
	end
	if playerDist <= chaseDist then
		player.arrestWatch = true
		if (c.travel == "backward" or playerDist <= nearDist) and player.policeWatch == "chase" then
			c.obj:queueLuaCommand("ai.setMode('chase')")
			c.obj:queueLuaCommand("ai.driveInLane('off')")
			player.policeWatch = "force"
		end
		if player.vel <= crawlSpeed then
			arrestPlayer()
		else
			player.arrestTimer = 0
		end
	else
		player.arrestWatch = false
		if c.travel == "backward" and player.policeWatch == "force" then
			c.obj:queueLuaCommand("ai.setMode('random')")
			c.obj:queueLuaCommand("ai.driveInLane('on')")
			player.policeWatch = "chase"
		end
	end
	if avoid == false then
		if player.policeWatch == "chase" then
			c.obj:queueLuaCommand("ai.setSpeed("..tostring(math.max(player.vel * ((readyDist / math.min(readyDist, playerDist) / 10) + 0.5), 5))..")") -- variable police speed based on distance
		end
		if player.policeWatch == "force" then
			c.obj:queueLuaCommand("ai.setSpeed("..tostring(math.max(player.vel * 1.5, 5))..")")
		end
	else
		c.obj:queueLuaCommand("ai.setSpeed("..tostring(speedingSpeed / 1.5)..")")
		c.obj:queueLuaCommand("ai.driveInLane('off')")
	end
end

local function onScenarioUIReady()
	player, traffic, barriers = {}, {}, {}
	player.obj = scenetree.findObject(playerName)
	player.damageTotal = playerSmash
	player.zone = 1
	player.timer = 0
	player.failTimer = 0
	player.policeWatch = "none"
	player.arrestWatch = false
	player.arrestTimer = 0
	player.crazy = 0
	player.success = true
	queue = 0
	roadblockZone = 1
	mainTimer = 0
	hornHold = 0
	zoneGap = 3
	roadblockMode = false
	for i = 1, #zoneData, 1 do
		zoneNames[i] = "spawnzone"..(i - 1)
	end
	local mapObjects = map.getTrackedObjects()
	local objects = {}
	for k, v in pairs(mapObjects) do
		local name = scenetree.findObjectById(k).name
		if string.find(name, trafficName) then
			table.insert(objects, name)
		end
	end
	table.sort(objects)
	for i, v in ipairs(objects) do
		local t = {}
		t.obj = scenetree.findObject(v)
		t.oldDamage = 0
		t.state = "normal"
		t.role = "normal"
		t.travel = "backward"
		t.far = 1
		t.timer = 0
		local pos = vec3(scenetree.findObject(zoneNames[i]):getPosition()) + vec3(0, 0, 0.5)
		local rot = scenetree.findObject(zoneNames[i]):getRotation()
		t.obj:setPositionRotation(pos.x, pos.y, pos.z, 0, 0, rot.z, rot.w)
		table.insert(traffic, t)
	end
	for i, v in ipairs(barrierNames) do
		local p = {}
		p.obj = scenetree.findObject(v)
		p.state = "none"
		p.role = "barrier"
		p.timer = 0
		table.insert(barriers, p)
	end
	for i, t in ipairs(traffic) do
		if string.find(t.obj.partConfig, "police") then -- AI police car
			t.obj:setPositionRotation(0, 0, 0, 0, 0, 0, 0)
			t.state = "none"
			t.role = "policeNormal"
		end
	end
end

local function onRaceStart()
	local policeVehicles = {"vehicles/coupe/police.pc", "vehicles/super/polizia.pc", "vehicles/etk800/police.pc", "vehicles/fullsize/police.pc", "vehicles/fullsize/bcpd.pc", "vehicles/etkc/police_highway_A.pc", "vehicles/autobello/carabinieri.pc", "vehicles/burnside/police.pc", "vehicles/sunburst/police.pc", "vehicles/sunburst/police_RS.pc", "vehicles/van/police.pc", "vehicles/roamer/sheriff.pc", "vehicles/roamer/sheriff_offroad.pc"}
	local crazyVehicles = {"ball", "barrels", "barrier", "blockwall", "bollard", "boxutility", "boxutility_large", "cannon", "caravan", "christmas_tree", "citybus", "cones", "dryvan", "flail", "flatbed", "flipramp", "haybale", "inflated_mat", "kickplate", "large_angletester", "large_bridge", "large_cannon", "large_crusher", "large_hamster_wheel", "large_metal_ramp", "large_roller", "large_spinner", "large_tilt", "metal_box", "metal_ramp", "piano", "pigeon", "roadsigns", "rocks", "rollover", "sawhorse", "semi", "streetlight", "suspensionbridge", "tanker", "tirestacks", "tirewall", "trafficbarrel", "tsfb", "tube", "van", "wall", "weightpad", "woodcrate", "woodplanks"}
	local ramplow = "vehicles/semi/t65_ramplow.pc"
	player.number = 1
	for i, v in ipairs(policeVehicles) do
		if player.obj.partConfig == policeVehicles[i] then
			player.crazy = 1
			player.number = 1.5 -- modifies speed limit, police start spawn time, police speed watch
		end
	end
	for i, v in ipairs(crazyVehicles) do
		if player.obj.jbeam == crazyVehicles[i] and player.crazy ~= 1 then
			player.crazy = 2
			player.number = 1.1
		end
	end
	if player.obj.partConfig == ramplow then
		player.crazy = 3
		player.number = 0.6
	end
	speedLimit = math.ceil(math.random(minSpeed, maxSpeed) / player.number) -- creates a new speed limit for all traffic cars; makes restarts more interesting but prevents traffic cars from bumping into each other
	for i, t in ipairs(traffic) do
		t.obj:queueLuaCommand("ai.setAggression("..tostring(trafficAggro)..")")
		t.obj:queueLuaCommand("ai.driveInLane('on')")
		t.obj:queueLuaCommand("ai.setSpeed("..tostring(speedLimit)..")")
		t.obj:queueLuaCommand("ai.setSpeedMode('limit')")
		t.obj:queueLuaCommand("ai.setMode('random')")
	end
end

local function onRaceTick(raceTickTime)
	mainTimer = mainTimer + raceTickTime
	player.pos = player.obj:getPosition()
	player.vel = map.objects[map.objectNames[player.obj.name]].vel:length()
	player.damage = map.objects[map.objectNames[player.obj.name]].damage
	if mainTimer < 1 then -- workaround for an odd bug where weight won't return properly at first
		player.obj:queueLuaCommand("obj:queueGameEngineLua('vWeight = ('..obj:calcBeamStats().total_weight..')')") -- creates a global variable with the player vehicle weight
	elseif mainTimer == 1 then
		if vWeight >= 2000 and funMode == false then
			player.damageTotal = math.ceil(player.damageTotal * (1 + (vWeight / 5000))) -- extra damage tolerance
		end
	end
	player.obj:queueLuaCommand("obj:queueGameEngineLua('vHorn = ('..electrics.values.horn..')')") -- creates a global variable for the horn state; 0 = off and 1 = on
	if vHorn == 1 then
		hornHold = hornHold + raceTickTime
		if hornHold == 1 and mainTimer <= 10 then
			timeLimit = 10000
			playerSmash = 1000000
			player.damageTotal = playerSmash
			funMode = true
			helper.flashUiMessage("Fun mode enabled!", 2, false)
		end
	end
	for i, t in ipairs(traffic) do
		t.pos = t.obj:getPosition()
		player.dist = (player.pos - t.pos):len()
		t.damage = map.objects[map.objectNames[t.obj.name]].damage
		if t.state == "set" then
			if t.timer >= 0.5 then -- any traffic timers, such as here, intentionally delay AI mode changes
				if t.role == "roadblock" then
					t.obj:queueLuaCommand("ai.setMode('stopping')")
					barriers[1].timer = 0
				else
					t.obj:queueLuaCommand("ai.setMode('random')")
				end
				t.obj:queueLuaCommand("ai.driveInLane('on')")
				t.state = "normal"
				t.far = 1
			end
			t.timer = t.timer + raceTickTime
		end
		if t.role == "roadblock" and (t.state == "normal" or t.state == "ready") then
			if barriers[1].timer < #barriers then -- uses barrier timer to delay barrier spawns for slight lag reduction
				for o, v in ipairs(barriers) do
					if (barriers[1].timer + 1) % 1 == 0 then
						if barriers[o].obj.name == barriers[barriers[1].timer + 1].obj.name then
							setPropSpawn(barriers[o], (o * 3.5) - 1.5)
						end
					end
				end
				barriers[1].timer = barriers[1].timer + raceTickTime
			end
		end
		if mainTimer >= policeTime * player.number then
			if t.role == "policeNormal" and player.policeWatch == "none" then
				t.state = "normal"
				player.policeWatch = "patrol"
			end
		end
		if player.dist <= nearDist then -- if you pass a traffic car, prepare it for teleportation
			t.state = "ready"
			t.far = 1
			if t.role == "policeNormal" then
				t.far = 2
			end
			t.timer = 0
		end
		if player.dist <= readyDist / 2 then -- a bunch of near area logic
			if t.role == "policeNormal" and player.policeWatch == "patrol" and (player.vel >= speedingSpeed * player.number or (player.dist <= 10 and t.damage >= t.oldDamage + (trafficSmash / 2))) then
				t.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
				t.obj:queueLuaCommand("ai.setAggression(1.2)")
				t.role = "policeChase"
				player.policeWatch = "chase"
			end
			if t.role == "policeChase" and player.dist <= 10 and t.damage >= policeSmash and roadblockMode == false then
				helper.flashUiMessage("The cops are really mad now!", 3, false)
				player.damageTotal = player.damageTotal * 1.5
				roadblockMode = true
			end
			if t.role == "roadblock" and player.dist <= nearDist then
				t.role = "policeChase"
			end
			t.oldDamage = t.damage
		end
		if player.zone > 1 then -- enables traffic spawning and other behaviors after you pass enough zones
			if player.dist >= farDist and t.state == "normal" then -- if you get really far away from a traffic car, prepare it for teleportation
				if t.role ~= "policeChase" then
					t.far = farMultiplier -- multiplier for ready distance; can prevent sudden vanishing
				end
				t.state = "ready"
				t.timer = 0
			end
			if queue > 0 then
				if player.dist >= (readyDist * t.far) and t.state == "ready" then
					if t.timer == 0 and t.zone == "none" then
						t.obj:queueLuaCommand("ai.setMode('disabled')")
						t.far = 0
					end
					if player.dist >= readyDist and t.timer >= 0.5 and t.obj.name == traffic[queue].obj.name then
						setSpawn(t)
						queue = -2 -- artificial delay for next spawn test; prevents lag and stuff
					end
					t.timer = t.timer + raceTickTime
				end
			end
		end
		if t.role == "policeChase" then -- function always runs if the police is chasing you
			chasePlayer(t)
		end
		if t.role ~= "policeChase" and t.damage >= trafficSmash * 2 then -- traffic cars try to stop driving if heavily damaged
			t.obj:queueLuaCommand("ai.setSpeed(0.01)")
		end
	end
	queue = queue + 1 -- this gets updated every race tick to prevent two cars from being teleported at the same time to the same spawn zone
	if queue > #traffic then
		queue = 1
	end
	if player.damage >= player.damageTotal then
		player.failTimer = player.failTimer + raceTickTime
		if player.success == true then
			bullettime.set(0.5)
			player.success = false
		end
		if player.failTimer >= 2 then
			if player.arrestWatch == true then
				scenario.finish({failed = failMessages.crashBusted})
			else
				scenario.finish({failed = failMessages.crash})
			end
		end
	end
	if timeLimit - 10 <= mainTimer then
		rushPlayer()
	end
end

local function onRaceResult()
	if player.success == true then
		if player.crazy == 1 then -- chaotic good
			scenario.finish({msg = winMessages.policeCar})
		elseif player.crazy == 2 then -- chaotic neutral
			scenario.finish({msg = winMessages.funnyCar})
		elseif player.crazy == 3 then -- chaotic evil
			scenario.finish({msg = winMessages.evilCar})
		elseif funMode == true then -- cheaty
			scenario.finish({msg = winMessages.normal})
		elseif mainTimer <= timeLimit - 60 then -- lawful evil
			scenario.finish({msg = winMessages.ultraFast})
		elseif mainTimer <= timeLimit - 40 then -- lawful neutral
			scenario.finish({msg = winMessages.veryFast})
		elseif mainTimer >= timeLimit - 5 then -- neutral good
			scenario.finish({msg = winMessages.almost})
		elseif roadblockZone > 1 then -- neutral evil
			scenario.finish({msg = winMessages.lawless})
		elseif player.policeWatch == "patrol" then -- lawful good
			scenario.finish({msg = winMessages.lawful})
		else -- true neutral
			scenario.finish({msg = winMessages.normal})
		end
	end
end

local function onBeamNGTrigger(data)
	if data.event == "enter" then
		if data.subjectName == player.obj.name then 
			if string.byte(tostring(data.triggerName)) == 115 then -- prevents stupid trigger confusion with bus stops
				for i, v in ipairs(zoneNames) do
					if data.triggerName == zoneNames[i] then
						if player.zone > i and i >= math.abs(zoneGap) + 1 then
							zoneGap = -3
						else
							zoneGap = 3
						end
						player.zone = i
					end
				end
				if player.zone + zoneGap > #zoneData then -- prevents spawn zone errors
					player.zone = 0
				end
			end
			if string.byte(tostring(data.triggerName)) == 102 then -- fail zones
				player.success = false
				if data.triggerName == "failzone0" then
					scenario.finish({failed = outMessages.roadClosed})
				end
				if data.triggerName == "failzone1" then
					scenario.finish({failed = outMessages.wrongWay})
				end
				if data.triggerName == "failzone2" then
					scenario.finish({failed = outMessages.hospital})
				end
			end
		end
		for i, t in ipairs(traffic) do
			if data.subjectName == t.obj.name then
				if string.byte(tostring(data.triggerName)) == 105 then -- intersections
					if t.role ~= "policeChase" then
						t.obj:queueLuaCommand("ai.setSpeed(8)")
					end
				end
				if string.byte(tostring(data.triggerName)) == 97 then -- action zones
					t.obj:queueLuaCommand("ai.setSpeed(0.5)")
				end
				for o, v in ipairs(zoneNames) do
					if data.triggerName == zoneNames[o] then
						t.zone = o
					end
				end	
				if data.triggerName == zoneNames[player.zone + zoneGap] then
					if t.role ~= "policeChase" then
						t.obj:queueLuaCommand("ai.setSpeed(10)") -- traffic cars slow down in zones where teleports are likely
					end
				end
				if data.triggerName == zoneNames[roadblockZone] then
					if t.role ~= "policeChase" then
						t.obj:queueLuaCommand("ai.setSpeed(0.5)") -- traffic cars slow down by police roadblocks
					end
				end
			end
		end
	end
	if data.event == "exit" then
		for i, t in ipairs(traffic) do
			if data.subjectName == t.obj.name then
				if string.byte(tostring(data.triggerName)) == 115 then
					if t.state ~= "set" then
						t.zone = "none"
					end
					if t.role ~= "policeChase" then
						t.obj:queueLuaCommand("ai.setSpeed("..tostring(speedLimit)..")")
					end
				end
			end
		end
	end
end

M.onScenarioUIReady = onScenarioUIReady
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult
M.onBeamNGTrigger = onBeamNGTrigger

return M